package App::bif::pull::project;
use strict;
use warnings;
use App::bif::Context;
use AnyEvent;
use Bif::Client;
use Coro;

our $VERSION = '0.1.0_26';

my $stderr;
my $stderr_watcher;

sub cleanup_errors {
    my $hub = shift;

    undef $stderr_watcher;
    $stderr->blocking(0);

    while ( my $line = $stderr->getline ) {
        print STDERR "$hub: $line";
    }

    return;
}

sub run {
    my $ctx = shift;
    $ctx->{no_pager}++;    # causes problems with something in Coro?
    $ctx = App::bif::Context->new($ctx);

    # Consider upping PRAGMA cache_size? Or handle that in Bif::Role::Sync?
    my $db = $ctx->dbw;

    my @locations = $db->get_hub_repos( $ctx->{hub} );
    $ctx->err( 'HubNotFound', 'hub not found: %s', $ctx->{hub} )
      unless @locations;

    my $hub = $locations[0];

    my @pinfo;
    foreach my $path ( @{ $ctx->{path} } ) {
        my $pinfo = $ctx->get_project( $path, $hub->{name} );

        return $ctx->err( 'ProjectNotFound', 'project not found: %s', $path )
          unless $pinfo;

        if ( $pinfo->{local} ) {
            print "Already imported: $pinfo->{path}\n";
            next;
        }

        push( @pinfo, $pinfo );
    }

    return $ctx->ok('PullProject')
      unless @pinfo;

    $|++;    # no buffering
    my $error;
    my $cv = AE::cv;

    my $client = Bif::Client->new(
        db            => $db,
        location      => $hub->{location},
        debug         => $ctx->{debug},
        debug_bifsync => $ctx->{debug_bifsync},
        on_error      => sub {
            $error = shift;
            $cv->send;
        },
    );

    $stderr = $client->child->stderr;

    $stderr_watcher = AE::io $stderr, 0, sub {
        my $line = $stderr->getline;
        if ( !defined $line ) {
            undef $stderr_watcher;
            return;
        }
        print STDERR "$hub->{name}: $line";
    };

    my $coro = async {
        eval {
            $db->txn(
                sub {
                    $ctx->update_localhub(
                        {
                            message =>
                              "pull project @{$ctx->{path}} $hub->{location}",
                        }
                    );

                    foreach my $pinfo (@pinfo) {
                        $db->xdo(
                            update => 'projects',
                            set    => 'local = 1',
                            where  => { id => $pinfo->{id} },
                        );
                    }

                    $client->on_update(
                        sub {
                            $ctx->lprint("$hub->{name}: $_[0]");
                        }
                    );

                    my $status = $client->sync_hub( $hub->{id} );

                    unless ( $status eq 'RepoMatch'
                        or $status eq 'RepoSync' )
                    {
                        $db->rollback;
                        $error = "unexpected status received: $status";
                        return cleanup_errors( $hub->{name} );

                    }

                    if ( $status eq 'RepoSync' ) {
                        $status = $client->transfer_hub_updates;

                        if ( $status ne 'TransferHubUpdates' ) {
                            $db->rollback;
                            $error = "unexpected status received: $status";
                            return cleanup_errors( $hub->{name} );
                        }
                    }

                    $status = $client->sync_projects;

                    unless ( $status eq 'ProjectSync' ) {
                        $db->rollback;
                        $error = "unexpected status received: $status";
                        return cleanup_errors( $hub->{name} );
                    }

                    $status = $client->transfer_project_related_updates;

                    if ( $status ne 'TransferProjectRelatedUpdates' ) {
                        $db->rollback;
                        $error = "unexpected status received: $status";
                        return cleanup_errors( $hub->{name} );
                    }
                    print "\n";

                    return cleanup_errors( $hub->{name} );
                }
            );
        };

        if ($@) {
            $error .= $@;
            print "\n";
        }

        $client->disconnect;
        return $cv->send( !$error );
    };

    if ( $cv->recv ) {
        $db->do('ANALYZE');
        return $ctx->ok('PullProject');
    }

    return $ctx->err( 'Unknown', $error );
}

1;
__END__

=head1 NAME

bif-pull-project -  import projects from a remote hub

=head1 VERSION

0.1.0_26 (2014-07-23)

=head1 SYNOPSIS

    bif pull project PATH... HUB

=head1 DESCRIPTION

The C<bif pull project> command imports projects from a hub. If a
project has been imported, then it is considered "local". If it has not
been imported then we call it "non-local". A project that we have
visibility of but have not registered the hub for we call "remote."

=head1 ARGUMENTS

=over

=item PATHS...

The path of the remote project to import.

=item HUB

The name of a previously registered hub.

=back

=head1 SEE ALSO

L<bif>(1)

=head1 AUTHOR

Mark Lawrence E<lt>nomad@null.netE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2013-2014 Mark Lawrence <nomad@null.net>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3 of the License, or (at your
option) any later version.

